"""
This file contains some low-level useful functions
"""

from __future__ import print_function

def write(*args,**kwargs):
    """
    Wrapper for print function or print to ensure compatibility with python 2
    The arguments are used similarly as the print_function
    They can also be generalized to python 2 cases
    """
    return print(*args,**kwargs)

def push_path(inp,*keys):
    """
    Follow in the dictionary inp the path indicated by the keys.
    If this path does not exists creates it.

    Args:
       inp (dict): dictionary
       keys (str): keys of the path to follow

    Returns:
       (``branch``,``key``) tuple, where

       * ``branch`` (dict): the dictionary of the second-last item of the path
       * ``key`` (str): the last item of the path

    Example:

       >>> inp={}
       >>> d,key=push_path(inp,'dft','nspin','mpol')
       >>> print (d,key)
       >>> print (inp)
       {},'mpol'
       {'dft': {'nspin': {}}}

       >>> inp={'dft': {'nspin': {'mpol': 2}}}
       >>> d,key=push_path(inp,'dft','nspin','mpol')
       >>> print (d,key)
       >>> print (inp)
       {'mpol': 2},'mpol'
       {'dft': {'nspin': {'mpol': 2}}}

    """
    tmp=inp
    for i,key in enumerate(keys):
        k=key
        if i==len(keys)-1: break
        tmp.setdefault(key,{})
        #if key not in tmp: tmp[key]={}
        tmp=tmp[key]
    return tmp,k


def dict_set(inp,*subfields):
    """Ensure the provided fields and set the value

    Provide a entry point to the dictionary.
    Useful to define a key in a dictionary that may not have the
    previous keys already defined.

    Arguments:
       inp (dict): the top-level dictionary
       subfields (str,object): keys, ordered by level, that have to be retrieved from topmost level of ``inp``.
          The last item correspond to the value to be set .

    Example:

       >>> inp={}
       >>> dict_set(inp,'dft','nspin','mpol',2)
       >>> print (inp)
       {'dft': {'nspin': {'mpol': 2}}}

    """
    if len(subfields) <= 1:
        raise ValueError('invalid subfields, the sequence should be longer than one item as the last one is the value to be given')
    keys=subfields[:-1]
    tmp,key=push_path(inp,*keys)
    tmp[key]=subfields[-1]

def dict_get(inp,*subfields):
    """Find the value of the provided sequence of keys in the dictionary, if available.

    Retrieve the value of the dictionary in a sequence of keys if it is available.
    Otherwise it provides as default value the last item of the sequence ``subfields``.

    Args:
       inp (dict): the top-level dictionary. Unchanged on exit.
       subfields (str,object): keys, ordered by level, that have to be retrieved from topmost level of ``inp``.
              The last item correspond to the value to be set.

    Returns:
       The value provided by the sequence of subfields if available, otherwise the default value given as the last item of the ``subfields`` sequence.

    """
    if len(subfields) <= 1:
        raise ValueError('invalid subfields, the sequence should be longer than one item as the last one is the value to be given')
    tmp=inp
    keys=subfields[:-1]
    val=subfields[-1]
    for key in keys:
        tmp=tmp.get(key)
        if tmp is None: return val
    return tmp

def sort_lists(sort_by,ascending,*lists):
    """
    Sort lists altogether following the lists indicated by the ``sort_by`` index.

    Args:

       sort_by (int):  the index of the list which has to be taken as reference for sorting
       ascending (bool): Sort is performed in ascending order if True

       *lists: sequence of lists to be mutually sorted. They have to be of the same length.

    Returns:
       tuple of sorted lists

    Example:
    >>> l1=[5,3,4]
    >>> l2=['c','t','q']
    >>> l3=[6,3,7]
    >>> print (sort_lists(0,True,l1,l2,l3))
    >>> print (sort_lists(2,True,l1,l2,l3))
    [(3, 4, 5), ('t', 'q', 'c'), (3, 7, 6)]
    [(3, 5, 4), ('t', 'c', 'q'), (3, 6, 7)]
    """
    import operator
    return zip(*sorted(zip(*lists), reverse= not ascending, key=operator.itemgetter(sort_by)))


def dict_merge(dest, src):
    """ Recursive dict merge. Inspired by :meth:`dict.update`, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``src`` is merged into
    ``dest``.  From :ref:`angstwad/dict-merge.py <https://gist.github.com/angstwad/bf22d1822c38a92ec0a9>`

    Arguments:
       dest (dict): dict onto which the merge is executed
       src (dict): dict merged into dest

    """
    import collections
    for k, v in src.items():
        if (k in dest and isinstance(dest[k], dict)
                and isinstance(src[k], collections.Mapping)):
            dict_merge(dest[k], src[k])
        else:
            dest[k] = src[k]

def file_time(filename):
    """
    Gives the date of the creation of the file, if exists.

    :param str filename: name of the file
    :returns: if the file exists, the date of the filename as per os.path.getmtime.
     Otherwise it returns 0
    """
    import os
    if os.path.isfile(filename):
        return os.path.getmtime(filename)
    else:
        return 0

def make_dict(inp):
    """
    Transform the instance ``inp`` into a python dictionary. If inp is already a dictionary, it perfroms a copy.

    Args:
       inp (dict): a instance of a Class which inherits from dict

    Returns:
       dict: the copy of the class, converted as a dictionary
    """
    import copy
    local_tmp=copy.deepcopy(inp)
    local_input={}
    local_input.update(local_tmp)
    return local_input

def function_signature_regenerator(target_kwargs_function,fun_name='',fun_docstring='',**kwargs):
    '''
    Generate the function of the name provided by `fun_name`, with signature provided by the
    kwargs dictionary.

    Args:
       target_kwargs_function (func): keyword arguments function that will be used for the generated function.
       fun_name (str): name of the regenerated function. If empty it will be the ``target_kwargs_functon.__name__`` prefixed by ``regenerated``, which will be copied in the docstring of the regenerated function.
       fun_docstring (str): docstring of the generated function, if empty it will take the docstring from ``target_kwargs_function``.
       **kwargs: keyword arguments which will represent the signature of the generated function.

    Example:
        >>> def write_kwargs(**kwargs):
        >>>     """
        >>>     Convert keyword arguments into a string
        >>>     """
        >>>     return str(kwargs)
        >>> write_opts=function_signature_regenerator(write_kwargs,fun_name='write_opts',opt1='default1',opt2='default2')
        >>> help(write_opts)
        >>> print (write_opts())
        Help on function write_opts:

        write_opts(opt1='default1', opt2='default2')
              Convert keyword arguments into a string

        {'opt1': 'default1', 'opt2': 'default2'}

    '''
    signature=option_line_generator(',',**kwargs).lstrip(',')
    docstring=target_kwargs_function.__doc__ if not fun_docstring else fun_docstring
    if docstring is None: docstring="Automatically generated function from the target function '"+target_kwargs_function.__name__+"'"
    docstring='   """\n'+docstring+'\n   """'
    fname="regenerated_"+target_kwargs_function.__name__ if  not fun_name else fun_name
    function="def %s(%s):\n%s\n   return target_function(**locals())" % (fname,signature,docstring)
    gen_locals={}
    gen_object=compile(function,'generated_fun','exec')
    eval(gen_object,{'target_function':target_kwargs_function},gen_locals)
    return gen_locals[fname]

def option_line_generator(separator='--',**kwargs):
    """
    Associate to each of the keyword arguments a command line argument.

    Args:
       separator (str): The string needed to separate the options.
       Might be '--' for command-line arguments, but also ',' for function signatures.

    Warning:
        The separator comes **before** the first argument therefore pay attention to
        lstrip it in case you want to use it as a function signature string.

    Example:
        >>> option_line_generator(arg1='val1',arg2='val2')
        '--arg1=val1 --arg2=val2'
    """
    command=''
    for option,value in kwargs.items():
        command+=separator+option+'="'+str(value)+'" '
    return command

def option_line_generator(separator='--',**kwargs):
    """
    Associate to each of the keyword arguments a command line argument.

    Args:
       separator (str): The string needed to separate the options.
       Might be '--' for command-line arguments, but also ',' for function signatures.

    Warning:
        The separator comes **before** the first argument therefore pay attention to
        lstrip it in case you want to use it as a function signature string.

    Example:
        >>> option_line_generator(arg1='val1',arg2='val2')
        '--arg1=val1 --arg2=val2'
    """
    command=''
    for option,value in kwargs.items():
        command+=separator+option+'="'+str(value)+'" '
    return command

def kw_pop(*args,**kwargs):
    """
    Treatment of kwargs. Eliminate from kwargs the tuple in args.
    """
    arg=kwargs.copy()
    key,default=args
    if key in arg:
        return arg,arg.pop(key)
    else:
        return arg,default


def find_files(regexp, archive=None):
    """
    Returns a list of the paths to the files that follow the regular expression
    regexp. They are searched from the current working directory or from an archive
    given as optional argument.


    :param regexp: A regular expression
    :type regexp: string
    :param archive: an opened tarfile archive (optional)
    :type archive:
    :returns: a list of all the paths that agree with the regexp
    :rtype: list of strings
    :raises: ValueError if the regexp does not find a single path.


    Example::

        #Find all python files in the current working directory
        find_files('*py')

        #An exmple outside of the current working directory
        find_files('*/log-*.yaml')

        #Example using a tarfile
        import tarfile
        my_archive = tarfile.open('archive.tar.gz')
        find_files('*/*/log-*.yaml', archive=my_archive)
    """
    import os

    #Get a list of all paths to files satisfying the regexp
    if archive is not None:
        paths = _find_files_from_archive(regexp, archive)
    else:
        paths = os.popen('ls '+regexp).read().splitlines()

    #Test that the regexp found files
    if paths == []:
        raise ValueError('The regexp "{}" leads to no file. '\
                         'Consider using another one.'.format(regexp))
    else:
        return paths


def _find_files_from_archive(re, archive):
    """
    This function retrieves the list of Logfiles instances
    from the file archived satisfying a regular expression.
    #function to identify an archive out of its regexp,
    #solves the bug in re for '*' (solved in Python 2.7.6)
    """
    import tarfile

    #Open the archive
    with tarfile.open(archive, 'r') as arch:
    #Return paths to logfiles satisfying the regexp
        return [f for f in arch.getnames()
                if all(pattern in f for pattern in re.split('*'))]

def ensure_copy(src,dest):
    """Copy src into dest.

    Guarantees that the file indicated by ``dest`` is a copy of the file ``src``

    Args:
      src (str): path of the source file. Should be valid.
      dest (src): path of the destination file

    Returns:
      bool: ``True`` if the file needed to be copied, ``False`` if ``src`` and ``dest`` are identical
    """
    import shutil,os
    copied=False
    if (os.path.isfile(dest) and os.stat(dest) != os.stat(src)) or not os.path.isfile(dest):
        shutil.copy2(src,os.path.dirname(dest))
        copied=True
    return copied

def ensure_dir(file_path):
    """
    Guarantees the existance on the directory given by the (relative) file_path

    Args:
       file_path (str): path of the directory to be created

    Returns:
       bool: True if the directory needed to be created, False if it existed already
    """
    import os
    directory = file_path
    created=False
    if not os.path.exists(directory):
        os.makedirs(directory)
        created=True
    return created

if __name__ == '__main__':
    import os

    #Tests of the find_files function
    #
    print("Test finding all python files in this directory")
    print(find_files("*py"))
    print()

    #
    print("Test finding the Utils.py file in this directory")
    print(find_files("Utils.py"))
    print()

    #
    print("Test raising a ValueError because the regexp leads to no files")
    try:
        find_files('*html')
    except ValueError as e:
        print('This raised the following ValueError:')
        print(e)
    print()

    #
    print("Test raising an exception because there is no such archive.")
    fname = 'file.tar.gz'
    if fname in os.popen('ls'): os.system('rm '+fname)
    #os.system('rm '+fname)
    try:
        find_files('*py', archive=fname)
    except Exception as e:
        #print(dir(e))
        print('This raised the following Exception:')
        print(e)
    print()

    #
    print("Test without error using an archive")
    os.system('find * -name "*py" | tar -zcvf '+fname+' -T -')
    os.system('ls '+fname)
    find_files('*py', archive=fname)
    os.system('rm '+fname)



"""
This module has the routines and data structures necessary to allow one to
generate visualizations of a atomic systems.

"""


class VSimVisualizer():
    def __init__(self, filename, xsize=600, ysize=600):
        from gi.repository import v_sim, GLib, Gtk
        self.win = v_sim.UiRenderingWindow.new(xsize, ysize, True, True)
        self.loop = GLib.MainLoop.new(None, False)
        self.win.connect_object('destroy', GLib.MainLoop.quit, self.loop)
        self.main = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
        self.main.add(self.win)
        self.main.show_all()
        self._set_file(filename)

    def _set_file(self, filename):
        from gi.repository import v_sim
        self.data = v_sim.DataAtomic.new(filename, None)
        self.data.load(0, None)
        self.win.getGlScene().setData(self.data)
        # v_sim.basic_parseConfigFiles()

    def show(self):
        # from gi.repository import GLib
        self.loop.run()

    def colorize_by_fragments(self):
        from gi.repository import v_sim
        self._scene_colorizer(v_sim, self.data, self.win.getGlScene())

    def _scene_colorizer(self, v_sim, data, scene):
        nodes = scene.getNodes()
        frag = data.getNodeProperties("Fragment")
        c = v_sim.DataColorizerFragment.new()
        c.setNodeModel(frag)
        nodes.pushColorizer(c)
        c.setActive(True)

    def colorizer_script(self, filename):
        import inspect
        towrite = ['scene = ' +
                   'v_sim.UiMainClass.getDefaultRendering().getGlScene()',
                   'data = scene.getData()']
        for line in inspect.getsource(self._scene_colorizer).split('\n')[1:]:
            towrite.append(line.lstrip(' '))
        f = open(filename, 'w')
        for line in towrite:
            f.write(line + '\n')
        f.close()


class VMDGenerator():
    """
    This class contains the routines you would use for visualization of
    a system using the VMD program.

    Attributes:
      representation (str): the vmd representation to draw with.
        https://www.ks.uiuc.edu/Research/vmd/allversions/repimages/#representations
      color (int): the default color to draw with.
    """
    def __init__(self, representation="CPK", color=16):
        self.representation = representation
        self.color = color

    def visualize_fragments(self, system, scriptfile, geomfile,
                            fragcolors=None):
        """
        This generates a script for visualizing the fragmentation of a
        system using VMD.

        Args:
          system (Fragments.System): the system to visualize.
          scriptfile (str): the name of the file to write the vmd script
            to (usually has extension .tcl)
          geomfile (str): the filename for where to write an xyz file
            of the system.
          fragcolors (dict): optionally, a dictionary from fragment ids to
            fragment colors. Colors are integers between 0 and 32.
        """
        from BigDFT.Fragments import Fragment
        from BigDFT.XYZ import XYZWriter

        # To create the XYZ file, we first make one big fragment.
        geomorder = Fragment()
        for fragid, frag in system.items():
            geomorder += frag

        # Then write it to file.
        with XYZWriter(geomfile, len(geomorder)) as ofile:
            for at in geomorder:
                ofile.write(at)

        # Get the matching so we can write the correct atom indices.
        matching = system.compute_matching(geomorder)

        # If fragcolors is not specified, we will generate it ourselves.
        if fragcolors is None:
            fragcolors = {}
            for i, s in enumerate(system):
                # 16 is black, which we will reserve.
                if i % 32 == 16:
                    c = str(32)
                else:
                    c = str(i % 32)
                fragcolors[s] = c

        # The header of the script file draws the whole system in black.
        outstr = self._get_default_header(geomfile)

        # This part colors individual fragments.
        modid = 1
        for fragid, frag in system.items():
            if fragid not in fragcolors:
                continue
            outstr += "mol addrep 0\n"
            outstr += """mol modselect """ + str(modid) + """ 0 index """
            outstr += " ".join([str(x) for x in matching[fragid]])
            outstr += "\n"
            outstr += """mol modcolor """
            outstr += str(modid) + """ 0 ColorID """ + \
                str(fragcolors[fragid]) + """\n"""
            modid += 1

        # Finally, write to file.
        with open(scriptfile, "w") as ofile:
            ofile.write(outstr)

    def visualize_qmmm(self, system, subsystem, target, scriptfile,
                       system_geomfile, subsystem_geomfile):
        """
        This routine generates a VMD script for visualizing a QM/MM setup.

        Args:
          system (Fragments.System): the full system to visualize.
          subsystem (Fragments.System): the qm part to visualize.
          target (str): the id of the target fragment of the QM/MM calculation.
          scriptfile (str): the name of the file to write the vmd script
            to (usually has extension .tcl)
          system_geomfile (str): the filename for where to write an xyz file
            of the system.
          subsystem_geomfile (str): the filename for where to write an xyz file
            of the subsystem.
        """
        from BigDFT.Fragments import Fragment
        from BigDFT.XYZ import XYZWriter

        # To create the XYZ file, we first make one big fragment.
        systemorder = Fragment()
        for fragid, frag in system.items():
            systemorder += frag
        subsystemorder = Fragment()
        for fragid, frag in subsystem.items():
            subsystemorder += frag

        # Then write it to file.
        with XYZWriter(system_geomfile, len(systemorder)) as ofile:
            for at in systemorder:
                ofile.write(at)
        with XYZWriter(subsystem_geomfile, len(subsystemorder)) as ofile:
            for at in subsystemorder:
                ofile.write(at)

        # The header of the script file draws the whole system in black.
        outstr = self._get_default_header(system_geomfile)

        # Draw the QM portion of the system
        matching = system.compute_matching(systemorder)
        modid = 1
        for fragid, frag in subsystem.items():
            if fragid == target:
                color = 4
            else:
                color = 0
            outstr += "mol addrep 0\n"
            outstr += """mol modselect """ + str(modid) + """ 0 index """
            outstr += " ".join([str(x) for x in matching[fragid]])
            outstr += "\n"
            outstr += """mol modcolor """
            outstr += str(modid) + """ 0 ColorID """ + str(color) + """\n"""
            modid += 1

        # Finally, write to file.
        with open(scriptfile, "w") as ofile:
            ofile.write(outstr)

    def _get_default_header(self, geomfile):
        outstr = """mol default style """+self.representation+"""\n"""
        outstr += """mol new """ + geomfile + "\n"
        outstr += """mol modcolor 0 0 ColorID """+str(self.color)+"""\n"""

        return outstr



"""Calculation datasets.

This module deals with the handling of series of calculations.
Classes and functions of this module are meant to simplify the approach to
ensemble calculations with the code, and to deal with parallel executions of
multiple instances of the code.

"""

from BigDFT.Calculators import Runner


def name_from_id(id):
    """Hash the id into a run name
    Construct the name of the run from the id dictionary

    Args:
        id (dict): id associated to the run

    Returns:
       str: name of the run associated to the dictionary ``id``
    """
    keys = list(id.keys())
    keys.sort()
    name = ''
    for k in keys:
        name += k + ':' + str(id[k]) + ','
    return name.rstrip(',')


def names_from_id(id):
    """Hash the id into a list of run names to search with the function id_in_names
       and add the separator ',' to have the proper value of a key
       (to avoid 0.3 in 0.39)
    """
    if id is None:
        return ['']
    else:
        return [name_from_id({k: v})+',' for k, v in id.items()]


class Dataset(Runner):
    """A set of calculations.

    Such class contains the various instances of a set of calculations with the
    code.
    The different calculations are labelled by parameter values and information
    that might then be retrieved for inspection and plotting.

    Args:
      label (str): The label of the dataset. It will be needed to identify the
          instance for example in plot titles or in the running directory.
      run_dir (str): path of the directory where the runs will be performed.
      input (dict): Inputfile to be used for the runs as default,
             can be overridden by the specific inputs of the run

    """

    def __init__(self, label='BigDFT dataset', run_dir='runs', **kwargs):
        """
        Set the dataset ready for appending new runs
        """
        from copy import deepcopy
        newkwargs = deepcopy(kwargs)
        Runner.__init__(self, label=label, run_dir=run_dir, **newkwargs)
        self.runs = []
        """
        List of the runs which have to be treated by the dataset these runs
        contain the input parameter to be passed to the various runners.
        """
        self.calculators = []
        """
        Calculators which will be used by the run method, useful to gather the
        inputs in the case of a multiple run.
        """

        self.results = {}
        """
        Set of the results of each of the runs. The set is not ordered as the
        runs may be executed asynchronously.
        """

        self.ids = []
        """
        List of run ids, to be used in order to classify and fetch the results
        """

        self.names = []
        """
        List of run names, needed for distinguishing the logfiles and
        input files. Each name should be unique to correctly identify a run.
        """

        self._post_processing_function = None

    def append_run(self, id, runner, **kwargs):
        """Add a run into the dataset.

        Append to the list of runs to be performed the corresponding runner and
           the arguments which are associated to it.

        Args:
          id (dict): the id of the run, useful to identify the run in the
             dataset. It has to be a dictionary as it may contain
             different keyword. For example a run might be classified as
             ``id = {'hgrid':0.35, 'crmult': 5}``.
          runner (Runner): the runner class to which the remaining keyword
             arguments will be passed at the input.

        Raises:
          ValueError: if the provided id is identical to another previously
             appended run.

        Todo:
           include id in the runs specification

        """
        from copy import deepcopy
        name = name_from_id(id)
        if name in self.names:
            raise ValueError('The run id', name,
                             ' is already provided, modify the run id.')
        self.names.append(name)
        # create the input file for the run, combining run_dict and input
        inp_to_append = deepcopy(self._global_options)
        inp_to_append.update(deepcopy(kwargs))
        # get the number of this run
        irun = len(self.runs)
        # append it to the runs list
        self.runs.append(inp_to_append)
        # append id and name
        self.ids.append(id)
        # search if the calculator already exists
        found = False
        for calc in self.calculators:
            if calc['calc'] == runner:
                calc['runs'].append(irun)
                found = True
                break
        if not found:
            self.calculators.append({'calc': runner, 'runs': [irun]})

    def process_run(self):
        """
        Run the dataset, by performing explicit run of each of the item of the
           runs_list.
        """
        self._run_the_calculations()
        return {}

    def _run_the_calculations(self, selection=None):
        for c in self.calculators:
            calc = c['calc']
            # we must here differentiate between a taskgroup run and a
            # separate run
            for r in c['runs']:
                if selection is not None and r not in selection:
                    continue
                inp = self.runs[r]
                name = self.names[r]
                self.results[r] = calc.run(name=name, **inp)

    def set_postprocessing_function(self, func):
        """Set the callback of run.

        Calls the function ``func`` after having performed the appended runs.

        Args:
           func (func): function that process the `inputs` `results` and
               returns the value of the `run` method of the dataset.
               The function is called as ``func(self)``.

        """
        self._post_processing_function = func

    def post_processing(self, **kwargs):
        """
        Calls the Dataset function with the results of the runs as arguments
        """
        if self._post_processing_function is not None:
            return self._post_processing_function(self)
        else:
            return self.results

    def fetch_results(self, id=None, attribute=None, run_if_not_present=True):
        """Retrieve some attribute from some of the results.

        Selects out of the results the objects which have in their ``id``
        at least the dictionary specified as input. May return an attribute
        of each result if needed.

        Args:
           id (dict): dictionary of the retrieved id. Return a list of the runs
               that have the ``id`` argument inside the provided ``id`` in the
               order provided by :py:meth:`append_run`.
           attribute (str): if present, provide the attribute of each of the
               results instead of the result object
           run_if_not_present (bool): If the run has not yet been performed
               in the dataset then perform it.

        Example:
           >>> study=Dataset()
           >>> study.append_run(id={'cr': 3},input={'dft':{'rmult':[3,8]}})
           >>> study.append_run(id={'cr': 4},input={'dft':{'rmult':[4,8]}})
           >>> study.append_run(id={'cr': 3, 'h': 0.5},
           >>>                  input={'dft':{'hgrids': 0.5, 'rmult':[4,8]}})
           >>> #append other runs if needed
           >>> #run the calculations (optional if run_if_not_present=True)
           >>> study.run()
           >>> # returns a list of the energies of first and the third result
           >>> # in this example
           >>> data=study.fetch_results(id={'cr': 3},attribute='energy')
        """
        names = names_from_id(id)
        fetch_indices = []
        selection_to_run = []
        for irun, name in enumerate(self.names):
            # add the separator ',' to have the proper value of a key
            # (to avoid 0.3 in 0.39)
            if not all([(n in name+',') for n in names]):
                continue
            if run_if_not_present and irun not in self.results:
                selection_to_run.append(irun)
            fetch_indices.append(irun)
        if len(selection_to_run) > 0:
            self._run_the_calculations(selection=selection_to_run)

        data = []
        for irun in fetch_indices:
            r = self.results[irun]
            data.append(r if attribute is None else getattr(r, attribute))
        return data

    def seek_convergence(self, rtol=1.e-5, atol=1.e-8, selection=None,
                         **kwargs):
        """
        Search for the first result of the dataset which matches the provided
        tolerance parameter. The results are in dataset order
        (provided by the :py:meth:`append_run` method) if `selection` is not
        specified.
        Employs the numpy :py:meth:`allclose` method for comparison.

        Args:
          rtol (float): relative tolerance parameter
          atol (float): absolute tolerance parameter
          selection (list): list of the id of the runs in which to perform the
               convergence search. Each id should be unique in the dataset.
          **kwargs: arguments to be passed to the :py:meth:`fetch_results`
               method.

        Returns:
          id,result (tuple): the id of the last run which matches the
                convergence, together with the result, if convergence is
                reached.

        Raises:
           LookupError: if the parameter for convergence were not found.
               The dataset has to be enriched or the convergence parameters
               loosened.
        """
        from numpy import allclose
        from futile.Utils import write
        to_get = self.ids if selection is None else selection

        id_ref = to_get[0]
        write('Fetching results for id "', id_ref, '"')
        ref = self.fetch_results(id=id_ref, **kwargs)
        ref = ref[0]
        for id in to_get[1:]:
            write('Fetching results for id "', id, '"')
            val = self.fetch_results(id=id, **kwargs)
            val = val[0]
            if allclose(ref, val, rtol=rtol, atol=atol):
                res = self.fetch_results(id=id_ref)
                label = self.get_global_option('label')
                write('Convergence reached in Dataset "' +
                      label+'" for id "', id_ref, '"')
                return (id_ref, res[0])
            ref = val
            id_ref = id
        raise LookupError('Convergence not reached, enlarge the dataset'
                          ' or change tolerance values')

    def get_times(self):
        from os import path as p
        from futile import Time as T
        time_files = {}
        run_dir = self.get_global_option('run_dir')
        time_files = [p.join(run_dir,
                             self.fetch_results(id=self.ids[c],
                                                attribute='data_directory')[0],
                             'time-' + self.names[c] + '.yaml'
                             ) for c in self.results
                      ]
        return T.TimeData(*time_files)

    def wait(self):
        from IPython.display import display, clear_output
        from aiida.orm import load_node
        running = len(self.results)
        while(running != 0):

            import time
            time.sleep(1)
            running = len(self.results)
            for c in self.results:
                pk = self.results[c]['node'].pk
                node = load_node(pk)
                if(node.is_finished):
                    running -= 1
                    # print(node.is_finished_ok)
            clear_output(wait=True)
            display(str(running)+" processes still running")

    def get_logfiles(self):
        logfiles = {}
        for c in self.results:
            try:
                logfiles[c] = self.calculators[0]['calc'].get_logs(
                    self.results[c]['node'].pk, self.names[c])
            except ValueError:
                logfiles[c] = self.results[c]
                print("no logfile for " + str(c))
        return logfiles


def combine_datasets(*args):
    """
    Define a new instance of the dataset class that should provide
    as a result a list of the runs of the datasets
    """
    full = Dataset(label='combined_dataset')
    # append the runs or each dataset
    for dt in args:
        for irun, runs in enumerate(dt.runs):
            calc = dt.get_runner(irun)
            id, dt.get_id(irun)
            full.append_run(id, calc, **runs)

    full.set_postprocessing_function(_combined_postprocessing_functions)


def _combined_postprocessing_functions(runs, results, **kwargs):
    pass



from matplotlib.axes import Axes
from matplotlib.widgets import AxesWidget

def show_image(imgfile,title=None):
    """
    Show image file using matplotlib imgread. Useful to bypass the 
    Jupyter bug for converting a notebook into a pdf file
    """
    import matplotlib.pyplot as plt
    import matplotlib.image as mpimg
    my_image = mpimg.imread(imgfile)
    fig = plt.figure()
    ax1 = fig.add_subplot(1,1,1)
    ax1.set_axis_off()
    ax1.imshow(my_image)
    if title: plt.title(title)
    plt.show()


#interesting slider for vertical matplotlib
#taken from https://stackoverflow.com/questions/25934279/add-a-vertical-slider-with-matplotlib
class VertSlider(AxesWidget):
    """
    A slider representing a floating point range.

    For the slider to remain responsive you must maintain a
    reference to it.

    Attributes:
      *ax*        : the slider :class:`matplotlib.axes.Axes` instance

      *val*       : the current slider value

      *hline*     : a :class:`matplotlib.lines.Line2D` instance
                     representing the initial value of the slider

      *poly*      : A :class:`matplotlib.patches.Polygon` instance
                     which is the slider knob

      *valfmt*    : the format string for formatting the slider text

      *label*     : a :class:`matplotlib.text.Text` instance
                     for the slider label

      *closedmin* : whether the slider is closed on the minimum

      *closedmax* : whether the slider is closed on the maximum

      *slidermin* : another slider - if not *None*, this slider must be
                     greater than *slidermin*

      *slidermax* : another slider - if not *None*, this slider must be
                     less than *slidermax*

      *dragging*  : allow for mouse dragging on slider

    Call :meth:`on_changed` to connect to the slider event
    """
    def __init__(self, ax, label, valmin, valmax, valinit=0.5, valfmt='%1.2f',
                 closedmin=True, closedmax=True, slidermin=None,
                 slidermax=None, dragging=True, **kwargs):
        """
        Create a slider from *valmin* to *valmax* in axes *ax*.

        Additional kwargs are passed on to ``self.poly`` which is the
        :class:`matplotlib.patches.Rectangle` which draws the slider
        knob.  See the :class:`matplotlib.patches.Rectangle` documentation
        valid property names (e.g., *facecolor*, *edgecolor*, *alpha*, ...).

        Parameters
        ----------
        ax : Axes
            The Axes to put the slider in

        label : str
            Slider label

        valmin : float
            The minimum value of the slider

        valmax : float
            The maximum value of the slider

        valinit : float
            The slider initial position

        label : str
            The slider label

        valfmt : str
            Used to format the slider value, fprint format string

        closedmin : bool
            Indicate whether the slider interval is closed on the bottom

        closedmax : bool
            Indicate whether the slider interval is closed on the top

        slidermin : Slider or None
            Do not allow the current slider to have a value less than
            `slidermin`

        slidermax : Slider or None
            Do not allow the current slider to have a value greater than
            `slidermax`


        dragging : bool
            if the slider can be dragged by the mouse

        """
        AxesWidget.__init__(self, ax)

        self.valmin = valmin
        self.valmax = valmax
        self.val = valinit
        self.valinit = valinit
        self.poly = ax.axhspan(valmin, valinit, 0, 1, **kwargs)

        self.hline = ax.axhline(valinit, 0, 1, color='r', lw=1)

        self.valfmt = valfmt
        ax.set_xticks([])
        ax.set_ylim((valmin, valmax))
        ax.set_yticks([])
        ax.set_navigate(False)

        self.connect_event('button_press_event', self._update)
        self.connect_event('button_release_event', self._update)
        if dragging:
            self.connect_event('motion_notify_event', self._update)
        self.label = ax.text(0.5, 1.03, label, transform=ax.transAxes,
                             verticalalignment='center',
                             horizontalalignment='center')

        self.valtext = ax.text(0.5, -0.03, valfmt % valinit,
                               transform=ax.transAxes,
                               verticalalignment='center',
                               horizontalalignment='center')

        self.cnt = 0
        self.observers = {}

        self.closedmin = closedmin
        self.closedmax = closedmax
        self.slidermin = slidermin
        self.slidermax = slidermax
        self.drag_active = False

    def _update(self, event):
        """update the slider position"""
        if self.ignore(event):
            return

        if event.button != 1:
            return

        if event.name == 'button_press_event' and event.inaxes == self.ax:
            self.drag_active = True
            event.canvas.grab_mouse(self.ax)

        if not self.drag_active:
            return

        elif ((event.name == 'button_release_event') or
              (event.name == 'button_press_event' and
               event.inaxes != self.ax)):
            self.drag_active = False
            event.canvas.release_mouse(self.ax)
            return

        val = event.ydata
        if val <= self.valmin:
            if not self.closedmin:
                return
            val = self.valmin
        elif val >= self.valmax:
            if not self.closedmax:
                return
            val = self.valmax

        if self.slidermin is not None and val <= self.slidermin.val:
            if not self.closedmin:
                return
            val = self.slidermin.val

        if self.slidermax is not None and val >= self.slidermax.val:
            if not self.closedmax:
                return
            val = self.slidermax.val

        self.set_val(val)

    def set_val(self, val):
        import six
        xy = self.poly.xy
        xy[1] = 0, val
        xy[2] = 1, val
        self.poly.xy = xy
        self.valtext.set_text(self.valfmt % val)
        if self.drawon:
            self.ax.figure.canvas.draw_idle()
        self.val = val
        if not self.eventson:
            return
        for cid, func in six.iteritems(self.observers):
            func(val)

    def on_changed(self, func):
        """
        When the slider value is changed, call *func* with the new
        slider position

        A connection id is returned which can be used to disconnect
        """
        cid = self.cnt
        self.observers[cid] = func
        self.cnt += 1
        return cid

    def disconnect(self, cid):
        """remove the observer with connection id *cid*"""
        try:
            del self.observers[cid]
        except KeyError:
            pass

    def reset(self):
        """reset the slider to the initial value if needed"""
        if (self.val != self.valinit):
            self.set_val(self.valinit)


def axis_from_data(fig,ax,data):
  "Transform a data tuple into axis coordinates"
  trans=fig.transFigure.inverted()
  ll=trans.transform(ax.transData.transform(data))
  return list(ll)

def data_from_data(fig,dst,src,data):
  "Transform a data tuple of anothe axis in the figure into data of another axis"
  trans=dst.transData.inverted()
  ll=trans.transform(src.transAxes.transform(data))
  return list(ll)

class AxisSet(Axes):
  def create_twin(self):
    self.twin=self.twinx()
  def cla(self):
    if hasattr(self,'twin'): self.twin.cla()
    super(AxisSet,self).cla()
  @classmethod
  def twinify(cls,ax):
    "Include the axis provided as a radix of the set"
    ax.__class__=cls
    ax.create_twin()

class FigureSet():
  """Container for multiple figures.

  Define a container for a plot with the possiblity to switch between simple and gnuplot plotting

  Arguments:
  title: The title of the master figure
  **kwargs: arguments for the axis instance

  """
  def __init__(self,**kwargs):
    import matplotlib.pyplot as plt
    from futile.Utils import kw_pop
    newkw,title=kw_pop('title','',**kwargs)    
    self.title=title
    self.figures=[]
    self.showing=False
    self.add(**newkw)

  def __getitem__(self,idx):
    if isinstance(idx,int):
      return self._get_figure(idx)
    else:
      return self._get_figure(figname=idx)

  def _locate(self,figname):
    for i,fig in enumerate(self.figures):
      if figname==fig['Title']: 
        return i,fig
    return None

  def exists(self,figname):
    "True if the Figure exists in the Set"
    return self._locate(figname) is not None

  def invoke(self,idx):
    import matplotlib.pyplot as plt
    "Invoke the Figure if it exists. Return None otherwise"
    if isinstance(idx,int):
      if idx >= len(self.figures): return None
      i=idx
    else:
      if not self.exists(idx): return None
      i,fig=self._locate(idx)
    figname=self.figures[i]['TrueTitle']
    return plt.figure(figname)

  def add(self,**kwargs):
    import matplotlib.pyplot as plt
    toadd=str(len(self.figures)+1)
    title=kwargs.get('title','Figure '+toadd if self.title is '' else self.title)
    if self.title != '' and len(self.figures)>0: 
      newtitle=title+' ('+self.title+')'
    else:
      newtitle=title
    quitbutton=kwargs.get('QuitButton',False)
    newfig=plt.figure(newtitle)
    axargs=kwargs.get('axes')
    #cfm=plt.get_current_fig_manager()
    totwin=False
    if 'twinaxes' in kwargs: totwin=kwargs.pop('twinaxes')
    if axargs is None:
      newax=newfig.add_subplot(111)
    else:
      rect=axargs.pop('rect')
      newax=plt.axes(rect,**axargs)
    if totwin: 
      AxisSet.twinify(newax)
    #connect the home key execpt for the first figure
    from functools import partial
    newfig.canvas.mpl_connect('key_press_event',partial(self._onkey_home,len(self.figures)))
    figdict={'Figure':newfig,'Axes':newax,'Title':title,'TrueTitle':newtitle}
    if quitbutton: figdict.update(self._put_quitbutton())
    self.figures.append(figdict)
    return newfig,newax

  def _onkey_home(self,ifig,event):
    import matplotlib.pyplot as plt
    number=event.key
    #print "pressed",number,'end'
    #return to the main figure if the key "home" is pressed
    if str(number)=='home': 
      #print 'raising window'
      self._raisewindow(0)
    elif number == 'q' or number == 'Q':
      if ifig==0: 
        self._quitall()
      else:
        figax=self.figures[ifig]
        plt.close(figax['Figure'])

  def _put_quitbutton(self):
    import matplotlib.pyplot as plt
    from matplotlib.widgets import Button
    quitbutton=Button(plt.axes([0.0, 0.0, 0.1, 0.075]), 'Quit')
    quitbutton.on_clicked(self._onclick_quitButton)
    return {'QuitButton':quitbutton}

  def _raisewindow(self,index):
    import matplotlib.pyplot as plt
    if self.invoke(index) is None: return
    cfm=plt.get_current_fig_manager()
    try:
      cfm.window.activateWindow()
      cfm.window.raise_()
    except:
      cfm.window.attributes('-topmost', True)
      #cfm.window.attributes('-topmost', False)
    return cfm

  def _quitall(self):
    import matplotlib.pyplot as plt
    for figax in self.figures:
      plt.close(figax['Figure'])
    self.showing=False
    #self.figures=[] #dereference everything
      
  def _onclick_quitButton(self,event):
    print ("Good bye!")
    self._quitall()

  def _get_figure(self,index=0,figname=None):
    if figname is not None:
      for fig in self.figures:
        if figname==fig['Title']: break
    else:
      fig=self.figures[index]
    fx=fig["Figure"]
    ax=fig["Axes"]
    return fx,ax

  def show(self,figname=None):
    import matplotlib.pyplot as plt
    if figname is not None and self.exists(figname): 
      fx,ax=self._get_figure(figname=figname)
      if not self.showing: self.show()
      fx.show()
    elif figname is None:
      self.showing=True
      plt.show()



"""
Command line interface (cli) for aiida_bigdft.

Register new commands either via the "console_scripts" entry point or plug them
directly into the 'verdi' command by using AiiDA-specific entry points like
"aiida.cmdline.data" (both in the setup.json file).
"""

from __future__ import absolute_import
import sys
import click
from aiida.cmdline.utils import decorators
from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.params.types import DataParamType


# See aiida.cmdline.data entry point in setup.json
@verdi_data.group('bigdft')
def data_cli():
    """Command line interface for aiida-bigdft"""
    pass


@data_cli.command('list')
@decorators.with_dbenv()
def list_():  # pylint: disable=redefined-builtin
    """
    Display all DiffParameters nodes
    """
    from aiida.orm import QueryBuilder
    from aiida.plugins import DataFactory
    DiffParameters = DataFactory('bigdft')

    qb = QueryBuilder()
    qb.append(DiffParameters)
    results = qb.all()

    s = ""
    for result in results:
        obj = result[0]
        s += "{}, pk: {}\n".format(str(obj), obj.pk)
    sys.stdout.write(s)


@data_cli.command('export')
@click.argument('node', metavar='IDENTIFIER', type=DataParamType())
@click.option('--outfile',
              '-o',
              type=click.Path(dir_okay=False),
              help='Write output to file (default: print to stdout).')
@decorators.with_dbenv()
def export(node, outfile):
    """Export a DiffParameters node (identified by PK, UUID or label) to plain text."""
    string = str(node)

    if outfile:
        with open(outfile, 'w') as f:
            f.write(string)
    else:
        click.echo(string)



"""Handle and automatize the parsing of input arguments

This module uses the same convention of the :f:mod:`yaml_parse` fortran module to define the command line arguments.
Such module can be used to define command line arguments of python scripts that follows the same conventions
or to generate python functions that have the same signature than the provided command arguments.

"""

def get_python_function(target_kwargs_function,func_name,func_spec):
    """Convert a argparse spec into a python function

    This function provides a python function with a signature indicated by the ``fun_spec`` dictionary
    With the conventions of the :f:mod:`yaml_argparse` modules.
    The :py:func:`futile.Utils.function_signature_regenerator` function is used for the conversion

    Args:
       target_kwargs_function (func): the keyword arguments function we want to give the signature to.
       func_name (str): Name of the function, usually the key of the dictionary whose ``func_spec`` is the value
       func_spec (dict) : dictionary of the function specifications to be provided to the
          :py:func:`futile.Utils.function_signature_regenerator` function.

    Returns:
       func: the genreated function with signature given by the arguments of ``func_spec``
         defaulting to their default value.

    Todo:
       Create the docstring of the generated function by also including the docstring of the arguments
    """
    from copy import deepcopy
    from futile.Utils import function_signature_regenerator as fsr
    fspec=deepcopy(func_spec)
    docstring=fspec.pop("help")
    if "shorthelp" in fspec: fspec.pop("shorthelp")

    key_args = {key: val["default"] for (key, val) in fspec["args"].items()}

    return fsr(target_kwargs_function, fun_name=func_name,
               fun_docstring=docstring,**key_args)



#!/usr/bin/env python

from __future__ import absolute_import
from setuptools import setup, find_packages
import os
import json
import sys
if __name__ == '__main__':
    # Provide static information in setup.json
    # such that it can be discovered automatically
    # create symlinks to subdirectories for easier use
    try:
      os.symlink('aiida_bigdft/PyBigDFT/BigDFT/', 'BigDFT')
    except:
      pass
    try:
      os.symlink('aiida_bigdft/futile/src/python/futile/','futile')
    except:
      pass
    with open('setup.json', 'r') as info:
        kwargs = json.load(info)
    setup(
        packages=find_packages(exclude=["aiida", "*.pyc", "aiida_bigdft/futile", "aiida_bigdft/PyBigDFT"]),
        # this doesn't work when placed in setup.json (something to do with str type)
        package_data={
            "": ["*"],
            # TODO be more specific with package data (but the line below isn't working)
            # "tests.input_files": ["*"],
        },
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        **kwargs)



# -*- coding: utf-8 -*-
"""
Data types provided by plugin

Register data types via the "aiida.data" entry point in setup.json.
"""

# You can directly use or subclass aiida.orm.data.Data
# or any other data type listed under 'verdi data'
from __future__ import absolute_import
from aiida.orm.nodes import Data
from voluptuous import Optional

from BigDFT import Logfiles
from BigDFT.scripts import InputGenerator

# A subset of BigDFT's command line options
cmdline_options = {
    Optional('ignore-case'): bool,
    Optional('ignore-file-name-case'): bool,
    Optional('ignore-tab-expansion'): bool,
    Optional('ignore-space-change'): bool,
    Optional('ignore-all-space'): bool,
}


class BigDFTParameters(Data):
    """
    Command line options for BigDFT.

    This class represents a python dictionary used to
    pass command line options to the executable.
    """

    # "voluptuous" schema  to add automatic validation
    # pylint: disable=redefined-builtin
    def __init__(self, dict=None, **kwargs):
        """
        Constructor for the data class

        Usage: ``BigDFTParameters(dict{'ignore-case': True})``

        :param parameters_dict: dictionary with commandline parameters
        :param type parameters_dict: dict

        """
#        dict = self.validate(dict)
        super(BigDFTParameters, self).__init__(**kwargs)
        self.set_attribute('dict', dict)

    def set_dict(self, dict):
        self.set_attribute('dict', dict)

    @property
    def dict(self):
        """
        Return the Logfile
        """
        return self.get_attribute('dict')

    def cmdline_params(self, file1_name, file2_name):
        """Synthesize command line parameters.

        e.g. [ '--ignore-case', 'filename1', 'filename2']

        :param file_name1: Name of first file
        :param type file_name1: str
        :param file_name2: Name of second file
        :param type file_name2: str

        """

    def __str__(self):
        """String representation of node.

        Append values of dictionary to usual representation. E.g.::

            uuid: b416cbee-24e8-47a8-8c11-6d668770158b (pk: 590)
            {'ignore-case': True}

        """
        string = super(BigDFTParameters, self).__str__()
        string += "\n" + str(self.get_attribute('dict'))
        return string

    def set_inputfile(hgrid, dico, init_input=None, units="reduced"):
        return InputGenerator.set_inputfile(hgrid, dico, init_input, units)

    def set_spin(name, nat):
        return InputGenerator.set_spin(name, nat)

    def set_kpoints(nat):
        return InputGenerator.set_kpoints(nat)

    def set_strain(strain, ngrids, dico):
        return InputGenerator.set_strain(strain, ngrids, dico)

    def set_restart():
        return InputGenerator.set_restart()

    def transform_to_orthorombic(dico):
        return InputGenerator.transform_to_orthorombic(dico)


class BigDFTLogfile(Data):
    """
    Wrapper around a Logfile object

    This class represents a BigDFT Logfile object.
    """
    def __init__(self, path=None, **kwargs):
        super(BigDFTLogfile, self).__init__(**kwargs)
        self.logfile = path

    @property
    def logfile(self):
        """
        Return the Logfile
        """
        return self.get_attribute('logfile')

    @logfile.setter
    def logfile(self, path):
        """
        Set the logfile

        :raise ValueError:
        """
        self.bigdftlogfile = Logfiles.Logfile(path)
        # replace forbidden chars in aiida dicts.
        if len(self.bigdftlogfile) > 0:
            logs = []
            for log in self.bigdftlogfile:
                try:
                    log.log['Timestamp of this run'] = \
                        log.log['Timestamp of this run'].strftime("%Y-%m-%d %H:%M:%S.%f")
                except KeyError:
                    pass
                logs.append(log.log)
            self.set_attribute('logfile', logs)
        else:
            self.bigdftlogfile.log['Timestamp of this run'] = \
                self.bigdftlogfile.log['Timestamp of this run'].strftime("%Y-%m-%d %H:%M:%S.%f")
            self.set_attribute('logfile', self.bigdftlogfile.log)

    def __str__(self):
        """String representation of node.

        Append values of dictionary to usual representation. E.g.::

            uuid: b416cbee-24e8-47a8-8c11-6d668770158b (pk: 590)
            {'ignore-case': True}

        """
        string = super(BigDFTLogfile, self).__str__()
        string += "\n" + str(self.get_attribute('logfile'))
        return string



# -*- coding: utf-8 -*-
"""
Parsers provided by aiida_bigdft.

Register parsers via the "aiida.parsers" entry point in setup.json.
"""
from __future__ import absolute_import

from aiida.engine import ExitCode
from aiida.parsers.parser import Parser
from aiida.plugins import CalculationFactory
from aiida.plugins import DataFactory
from aiida.common.exceptions import ValidationError

BigDFTCalculation = CalculationFactory('bigdft')
BigDFTLogfile = DataFactory('bigdft_logfile')


class BigDFTParser(Parser):
    """
    Parser class for parsing output of calculation.
    """

    def __init__(self, node):
        """
        Initialize Parser instance

        Checks that the ProcessNode being passed was produced
        by a BigDFTCalculation.

        :param node: ProcessNode of calculation
        :param type node: :class:`aiida.orm.ProcessNode`
        """
        from aiida.common import exceptions
        super(BigDFTParser, self).__init__(node)
        if not issubclass(node.process_class, BigDFTCalculation):
            raise exceptions.ParsingError("Can only parse BigDFTCalculation")

    def parse(self, **kwargs):
        """
        Parse outputs, store results in database.

        :returns: an exit code, if parsing fails
        (or nothing if parsing succeeds)
        """
        output_filename = self.node.get_option('output_filename')
        jobname = self.node.get_option('jobname')
        if jobname is not None:
            output_filename = "log-" + jobname + ".yaml"
        # Check that folder content is as expected
        files_retrieved = self.retrieved.list_object_names()
        files_expected = [output_filename]
        # Note: set(A) <= set(B) checks whether A is a subset of B
        if not set(files_expected) <= set(files_retrieved):
            self.logger.error("Found files '{}', expected to find '{}'".format(
                files_retrieved, files_expected))
            return self.exit_codes.ERROR_MISSING_OUTPUT_FILES

        # add output file
        self.logger.info("Parsing '{}'".format(output_filename))
#        print(self.retrieved._repository._get_base_folder().get_abs_path(output_filename))
        output = BigDFTLogfile(self.retrieved._repository._get_base_folder().
                               get_abs_path(output_filename))
        try:
            output.store()
        except ValidationError:
            self.logger.info("Impossible to store LogFile - ignoring '{}'".
                             format(output_filename))

#        with self.retrieved.open(output_filename, 'rb') as handle:
#            output_node = SinglefileData(file=handle)
#        output_dict_aiida=orm.Dict(dict=output_dict)
#        output_dict_aiida.store()
#        output_log_aiida=BigDFTLogfile(output)
        self.out('bigdft_logfile', output)

        return ExitCode(0)



from BigDFT.Calculators import SystemCalculator, Runner
from futile.Utils import write as safe_print
from futile.Utils import dict_merge

import six
import os

from aiida.common import datastructures
from aiida.engine import launch, CalcJob
from aiida.orm.nodes.data import List, SinglefileData
from aiida.orm import load_code, load_node


class BigDFTCalcJob(CalcJob):
    """Simple `CalcJob` implementation"""

    @classmethod
    def define(cls, spec):
        super(BigDFTCalcJob, cls).define(spec)
        spec.input('metadata.options.command_line',
                   valid_type=six.string_types, default='')
        spec.input('metadata.options.local_copy_list', valid_type=List)
        spec.input('metadata.options.retrieve_list', valid_type=List)

    def prepare_for_submission(self, folder):
        codeinfo = datastructures.CodeInfo()
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.withmpi = self.inputs.metadata.options.withmpi
        if self.inputs.metadata.options.command_line != '':
            codeinfo.cmdline_params = \
                self.inputs.metadata.options.command_line.split()
        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.local_copy_list = \
            self.inputs.metadata.options.local_copy_list.get_list()
        calcinfo.retrieve_list = \
            self.inputs.metadata.options.retrieve_list.get_list()
        return calcinfo


class AiidaCalculator(SystemCalculator):
    """Run of an aiida process.
    """

    def __init__(self,
                 omp=os.environ.get('OMP_NUM_THREADS', '1'),
                 mpi_run=os.environ.get('BIGDFT_MPIRUN', ''),
                 dry_run=False,
                 skip=False,
                 code="bigdft@localhost",
                 verbose=True,
                 **kwargs):
        Runner.__init__(self, omp=str(omp), mpi_run=mpi_run,
                        dry_run=dry_run, skip=skip, verbose=verbose, **kwargs)
        # Build the command setting the number of omp threads
        self.code = load_code(code)
        self.job = BigDFTCalcJob
        # we don't want the executable and such
        self.command = ""
        # ignore this for datasets, as aiida splits each
        # computation in its own folder
        self.run_dir = "."
        # store logfiles and names in order to
        # skip execution if necessary.
        self.logfiles = {}
        # store output information on all runs.
        self.outputs = {}
        safe_print('Initialize an Aiida Calculator for %s with %d machine(s)' %
                   ((self.code),
                    self._global_options.get('num_machines', 1)),
                   ', %d processes per machine, and %s cores per process' %
                   ((self._global_options.get('mpiprocs_per_machine', 1),
                    self._global_options['omp'])))

    def _ensure_run_directory(self):
        pass

    def pre_processing(self):
        """
        Generate files for run
        """
        self.metadata = {
            'label': self.run_options.get('name', ''),
            'description': "AiiDA calc" + self.run_options.get('name', ''),
            'options': {
                'resources': {
                    'num_machines': self.run_options.get('num_machines', 1),
                    'num_mpiprocs_per_machine':
                    self.run_options.get('mpiprocs_per_machine', 1),
                    'num_cores_per_mpiproc': self.run_options.get('omp', 1)
                },
                'withmpi': True,
                'environment_variables': {
                    "OMP_NUM_THREADS": str(self.run_options.get('omp', 1))
                },
                'max_wallclock_seconds': 
                self.run_options.get('walltime', 30 * 60),
                'scheduler_stdout': '_scheduler-stdout.txt',
                'scheduler_stderr': '_scheduler-stderr.txt'
            }
        }
        if 'queue_name' in self.run_options:
            self.metadata['options']['queue_name'] = \
                self.run_options['queue_name']
        if 'account' in self.run_options:
            self.metadata['options']['account'] = self.run_options['account']
        if 'mem' in self.run_options:
            self.metadata['options']['max_memory_kb'] = self.run_options['mem']
        if 'qos' in self.run_options:
            self.metadata['options']['qos'] = self.run_options['qos']
        if 'metadata' in self.run_options:
            self.metadata['options'].update(self.run_options['metadata'])
#        print(self.metadata)
        run_args = SystemCalculator.pre_processing(self)

        # input files needed for calculation.
        # This is basically prepare_for_submission
        input_filedata = SinglefileData(
            file=os.path.abspath(self._get_inputfilename())).store()

        local_copy_list = [
            (input_filedata.uuid, 
             input_filedata.filename, 
             input_filedata.filename)]
        # setup pseudopotentials if needed
        for filename in self._get_inputpseudos():
            pseudo_filedata = SinglefileData(
                file=os.path.abspath(filename)).store()
            local_copy_list.append(
                (pseudo_filedata.uuid, pseudo_filedata.filename, filename))
        posinp_filename = self.run_options['input'].get('posinp', None)
        if posinp_filename is None:
            posinp_filename = self.run_options.get('posinp', None)
        if os.path.isfile(str(posinp_filename)):
            posinp_filedata = SinglefileData(
                file=os.path.abspath(posinp_filename)).store()
            local_copy_list.append(
                (posinp_filedata.uuid,
                 posinp_filedata.filename,
                 posinp_filename))
        name = self.run_options.get('name', 'log.yaml')
        outfile = name
        timefile = "time.yaml"
        if name != 'log.yaml':
            outfile = "log-"+name+".yaml"
            timefile = "time-"+name+".yaml"
#        self.job.local_options={
        retrieve_list = List()
        retrieve_list.extend(
            [outfile, 
             timefile, 
             "forces*", 
             "final*", 
             ["./debug/bigdft-err*", ".", 2]])
        retrieve_list.store()
        self.metadata['options']['retrieve_list'] = retrieve_list
        local_copy_List = List()
        local_copy_List.extend(local_copy_list)
        local_copy_List.store()
        self.metadata['options']['local_copy_list'] = local_copy_List
        self.metadata['options']['command_line'] = self._get_command()
#        }
        return run_args

    def process_run(self, **kwargs):
        """Finally launch the code.
        Routine associated to the running of the ``bigdft`` executable.
        """
        # check if the debug file will be updated (case of erroneous run)
        timedbg = self._get_debugfile_date()
        # Run the command
        out = launch.run(self.job, code=self.code, metadata=self.metadata)
        name = self.run_options.get('name', '')
        self.outputs[name] = out
        logname = out['retrieved']._repository._get_base_folder(
        ).get_abs_path(self._get_logname())
        return {'timedbg': timedbg, 'logname': logname}

    def run(self, **kwargs):
        self._run_options(**kwargs)
        name = self.run_options.get('name', '')
        asynchronous = self.run_options.get('async', False)
        if self.run_options['skip'] and name in self.logfiles:
            return self.logfiles[name]
        else:
            run_args = self.pre_processing()
            if(asynchronous):
                node = launch.submit(
                    self.job, code=self.code, metadata=self.metadata)
                run_results = {'node': node}
                dict_merge(dest=run_args, src=run_results)
                return run_args
            else:
                run_results = self.process_run(**run_args)
    #          safe_print('run_args',run_args,'run_results',run_results)
                dict_merge(dest=run_args, src=run_results)
    #          safe_print('run_updated, again',run_args)
                self.logfiles[name] = SystemCalculator.post_processing(
                    self, **run_args)
                print("setting data dir to " +
                      self.outputs[name]
                      ['retrieved']._repository._get_base_folder().abspath)
                setattr(self.logfiles[name], "data_directory", 
                        self.outputs[name]
                        ['retrieved']._repository._get_base_folder().abspath)
                return self.logfiles[name]

    def submit(self, **kwargs):
        self._run_options(**kwargs)
        name = self.run_options.get('name', '')
        if self.run_options['skip'] and name in self.logfiles:
            return self.logfiles[name]
        else:
            run_args = self.pre_processing()
            node = launch.submit(self.job, code=self.code,
                                 metadata=self.metadata)
            run_results = {'node': node}
            dict_merge(dest=run_args, src=run_results)
            return run_args

    def wait(self):
        from IPython.display import display, clear_output
        running = len(self.results)
        while(running != 0):

            import time
            time.sleep(1)
            running = len(self.results)
            for c in self.results:
                pk = self.results[c]['node'].pk
                node = load_node(pk)
                if(node.is_finished):
                    running -= 1
                    # print(node.is_finished_ok)
            clear_output(wait=True)
            display(str(running)+" processes still running")

    def get_logs(self, pk, name):
        #      name = self.run_options.get('name', '')
        self.outputs[name] = load_node(pk).outputs
        logname = 'log-' + name + '.yaml' if name else 'log.yaml'
        logfile = self.outputs[name]['retrieved']._repository._get_base_folder(
        ).get_abs_path(logname)
        # to avoid comparing times with wrong input name
        self.run_options['name'] = name
        timedbg = self._get_debugfile_date()
        command = self._get_command()
        run_args = {'timedbg': timedbg, 'logname': logfile, 'command': command}
        self.logfiles[name] = SystemCalculator.post_processing(
            self, **run_args)
        print("setting data dir to " +
              self.outputs[name]
              ['retrieved']._repository._get_base_folder().abspath)
        setattr(self.logfiles[name], "data_directory", self.outputs[name]
                ['retrieved']._repository._get_base_folder().abspath)
        return self.logfiles[name]

    def get_time(self):
        from os import path as p
        from futile import Time as T
        from importlib import reload
        reload(T)
        datadir = self.fetch_results(attribute='data_directory')
        run_dir = self.get_global_option('run_dir')
        time_files = [p.join(run_dir, d, 'time-'+n+'.yaml')
                      for d, n in zip(datadir, self.names)]
        return T.TimeData(*time_files)

    def _get_inputpseudos(self):
        return {}



# -*- coding: utf-8 -*-

"""
Calculations provided by aiida_bigdft.

Register calculations via the "aiida.calculations" entry point in setup.json.
"""
from __future__ import absolute_import

import six
import os

from aiida import orm
from aiida.common import datastructures, exceptions
from aiida.engine import CalcJob
from aiida.orm.nodes.data import List, SinglefileData, Float, Dict
from aiida.plugins import DataFactory

from BigDFT import Calculators as BigDFT_calc
from BigDFT import InputActions as BigDFT_input
from BigDFT import Inputfiles as BigDFT_files
BigDFTParameters = DataFactory('bigdft')
BigDFTLogfile = DataFactory('bigdft_logfile')

# just override SystemCalculator constructor to avoid checking BIGDFT_ROOT variable


class PluginSystemCalculator(BigDFT_calc.SystemCalculator):
    def __init__(self,
                 omp=os.environ.get('OMP_NUM_THREADS', '1'),
                 mpi_run=os.environ.get('BIGDFT_MPIRUN', ''),
                 dry_run=False, skip=False, verbose=True):
        # Use the initialization from the Runner class (so all options inside
        # __global_options)
        BigDFT_calc.Runner.__init__(self, omp=str(omp), mpi_run=mpi_run,
                                    dry_run=dry_run, skip=skip, verbose=verbose)
        self.command = ""


class BigDFTCalculation(CalcJob):
    """
    AiiDA calculation plugin wrapping the BigDFT python interface.
    """
    _POSINP_FILE_NAME = 'posinp.xyz'
    _INPUT_FILE_NAME = 'input.yaml'
    _OUTPUT_FILE_NAME = 'log.yaml'
    _TIMING_FILE_NAME = 'time.yaml'

    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        # yapf: disable
        super(BigDFTCalculation, cls).define(spec)
        spec.input('metadata.options.resources', valid_type=dict, default={
                   'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        spec.input('metadata.options.parser_name',
                   valid_type=six.string_types, default='bigdft')
        spec.input('metadata.options.output_filename',
                   valid_type=six.string_types, default=cls._OUTPUT_FILE_NAME)
        spec.input('metadata.options.jobname',
                   valid_type=six.string_types, required=False)
        spec.input('parameters', valid_type=BigDFTParameters,
                   help='Command line parameters for BigDFT')
        spec.input('structure', valid_type=orm.StructureData,
                   help='StructureData struct')
        spec.input('structurefile', valid_type=orm.Str, help='xyz file',
                   default=lambda: orm.Str(cls._POSINP_FILE_NAME))
        spec.input('pseudos', valid_type=List, help='',
                   default=lambda: List(), required=False)
        spec.input('kpoints', valid_type=Dict, help='kpoint mesh or kpoint path',
                   default=lambda: Dict(dict={}), required=False)
        spec.input('extra_retrieved_files', valid_type=List,
                   help='', default=lambda: List())
        spec.output('bigdft_logfile', valid_type=BigDFTLogfile,
                    help='BigDFT log file as a dict')
        spec.exit_code(100, 'ERROR_MISSING_OUTPUT_FILES',
                       message='Calculation did not produce all expected output files.')

    def prepare_for_submission(self, folder):
        """
        Create input files.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily place all files needed by
            the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """
        # check that either structure or structurefile are set
        # if both are set, fail.
#        if self.inputs.structurefile is None and self.inputs.structure is None:
#            raise exceptions.InputValidationError("either structure or structurefile must be set")
#        if self.inputs.structurefile is not None and self.inputs.structure is not None:
#            raise exceptions.InputValidationError("Only one of structure or structurefile must be set")

        # somehow aiida sends back unicode strings here
        dico = BigDFT_files.Inputfile()
        dico.update(self.inputs.parameters.dict)

        bigdft_calc = PluginSystemCalculator()
        local_copy_list = []
        # if the structure is not already inside input dict
        if not dico["posinp"]:
            posinp_filename = self.inputs.structurefile.value
            if self.inputs.structure is not None:
                print("writing input posinp file")
                posinp_string = self.inputs.structure._prepare_xyz()[0]
                if "jobname" not in self.inputs.metadata.options:
                    posinp_filename = self._POSINP_FILE_NAME
                else:
                    posinp_filename = self.inputs.metadata.options.jobname + ".xyz"
    #            bigdft_calc.update_global_options(units="angstroem")
                posinp_file = open(posinp_filename, "wb")

                posinp_file.write(posinp_string)
                posinp_file.close()
            posinp_filedata = SinglefileData(
                file=os.path.abspath(posinp_filename)).store()

    #        BigDFT_input.set_atomic_positions(dico, posinp_filename)
    #        bigdft_calc._run_options(posinp={})

            local_copy_list = [
                (posinp_filedata.uuid, posinp_filedata.filename, posinp_filedata.filename)]

        # setup pseudopotentials if needed
        if self.inputs.pseudos is not None:
            for filename in self.inputs.pseudos:
                pseudo_filedata = SinglefileData(
                    file=os.path.abspath(filename)).store()
                local_copy_list.append(
                    (pseudo_filedata.uuid, pseudo_filedata.filename, filename))

        if self.inputs.kpoints:
            var = {}
            var["kpt"] = self.inputs.kpoints.dict
            dico.update(var)

        # generate yaml input file from dict and whatever

        if "jobname" in self.inputs.metadata.options:
            bigdft_calc.update_global_options(
                name=self.inputs.metadata.options.jobname)

        bigdft_calc._run_options(input=dico)
        bigdft_calc.pre_processing()
        if "jobname" not in self.inputs.metadata.options:
            input_filename = self._INPUT_FILE_NAME
        else:
            input_filename = self.inputs.metadata.options.jobname + ".yaml"
        input_filedata = SinglefileData(
            file=os.path.abspath(input_filename)).store()
        local_copy_list.append(
            (input_filedata.uuid, input_filedata.filename, input_filename))

        codeinfo = datastructures.CodeInfo()
        codeinfo.code_uuid = self.inputs.code.uuid
        outfile = self.inputs.metadata.options.output_filename
        timefile = self._TIMING_FILE_NAME
        if "jobname" in self.inputs.metadata.options:
            outfile = "log-" + self.inputs.metadata.options.jobname + ".yaml"
            timefile = "time-" + self.inputs.metadata.options.jobname + ".yaml"

#        codeinfo.stdout_name = outfile
        codeinfo.withmpi = self.inputs.metadata.options.withmpi
        if "jobname" in self.inputs.metadata.options:
            codeinfo.cmdline_params = ["--name=" +
                                       self.inputs.metadata.options.jobname]
        #local_copy_list = []
        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.local_copy_list = local_copy_list
        calcinfo.retrieve_list = [outfile,
                                  timefile,
                                  "forces_posinp.yaml",
                                  "final_posinp.yaml",
                                  ["./debug/bigdft-err*", ".", 2]]
        calcinfo.retrieve_list.extend(self.inputs.extra_retrieved_files)
        return calcinfo



"""
Relax workchain.

"""
from aiida import orm
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import WorkChain, append_, ToContext
from aiida.orm.nodes.data import List
from aiida.plugins import WorkflowFactory, DataFactory
from BigDFT import InputActions
from futile import YamlIO

import numpy
BigDFTBaseWorkChain = WorkflowFactory('bigdft')
BigDFTParameters = DataFactory('bigdft')
StructureData = DataFactory('structure')
ArrayData = DataFactory('array')


class BigDFTRelaxWorkChain(WorkChain):
    """Structure relaxation workchain."""
    _verbose = False

    @classmethod
    def define(cls, spec):
        super(BigDFTRelaxWorkChain, cls).define(spec)
        spec.expose_inputs(BigDFTBaseWorkChain,
                           exclude=['parameters',
                                    'extra_retrieved_files'])
        spec.input('parameters', valid_type=BigDFTParameters, required=False,
                   default=lambda: orm.Dict(), help='param dictionary')
        spec.input('extra_retrieved_files', valid_type=List, required=False,
                   help='', default=lambda: List())
        spec.input('relax.perform', valid_type=orm.Bool, required=False,
                   default=lambda: orm.Bool(True), help='perform relaxation')
        spec.input('relax.algo', valid_type=orm.Str,
                   default=lambda: orm.Str('FIRE'),
                   help='algorithm to use during relaxation')
        spec.input('relax.threshold_forces', valid_type=orm.Float, required=False,
                   default=lambda: orm.Float(0.0), help='energy cutoff value, in ev/Ang')
        spec.input('relax.steps', valid_type=orm.Int, required=False,
                   default=lambda: orm.Int(50),
                   help='number of relaxation steps to perform.')
        spec.outline(
            cls.relax,
            cls.results,
        )
        spec.expose_outputs(BigDFTBaseWorkChain)
        spec.output('relaxed_structure', valid_type=StructureData,
                    required=False)
        spec.output('forces', valid_type=ArrayData, required=False)
        spec.output('total_energy', valid_type=orm.Float, required=False)
        spec.exit_code(101, 'ERROR_FAILED_RELAX',
                       'Subprocess failed for relaxation')

    def relax(self):

        self.ctx.inputs = AttributeDict(
            self.exposed_inputs(BigDFTBaseWorkChain))
        inputdict = self.inputs.parameters.dict
        if self.inputs.relax.perform:
            InputActions.optimize_geometry(inputdict,
                                           self.inputs.relax.algo.value,
                                           self.inputs.relax.steps.value)
        if self.inputs.relax.threshold_forces is not None:
            InputActions.dict_set(inputdict, 'geopt', 'forcemax',
                                  self.inputs.relax.threshold_forces.value / 0.52917721067121)

        self.ctx.inputs.parameters = BigDFTParameters(dict=inputdict)

        # gather outputs
        extra_retrieved_files = List()
        if self.inputs.extra_retrieved_files is not None:
            extra_retrieved_files.set_list(
                self.inputs.extra_retrieved_files.get_list())
        extension = "xyz"
        posinp = self.inputs.parameters.dict.get('posinp')
        if posinp is not None:
            try:
                extension = posinp['properties']['format']
            except KeyError:
                extension = "yaml"
        extra_retrieved_files.extend([["./data*/*." + extension, ".", 2],
                                      ["./data*/geopt.mon", ".", 2],
                                      ["./final_*", ".", 2]])
        self.ctx.inputs.extra_retrieved_files = extra_retrieved_files

        node = self.submit(BigDFTBaseWorkChain, **self.ctx.inputs)
        return ToContext(work=append_(node))

    def results(self):
        workchain = self.ctx.work[-1]

        if not workchain.is_finished_ok:
            self.report(
                'Relaxation failed with exit status {}'.
                format(workchain.exit_status))
            return self.exit_codes.ERROR_FAILED_RELAX
        extension = "xyz"
        posinp = self.inputs.parameters.dict.get('posinp')
        if posinp is not None:
            try:
                extension = posinp['properties']['format']
            except KeyError:
                extension = "yaml"
        if self.inputs.relax.perform:
            outstruct = "final_posinp." + extension
            repo = workchain.outputs.retrieved._repository._get_base_folder()
            if "jobname" in self.inputs.run_opts.get_dict()['options']:
                outstruct = "final_" +\
                    self.inputs.run_opts.get_dict(
                    )['options']['jobname'] + ".xyz"
            try:
                sf = repo.get_abs_path(outstruct, check_existence=True)
            except OSError:
                # final_*xyz file not found, it did not finish.
                # We can output last *xyz file in data folder, restart, or fail
                # do we have posout files ?
                subname = "data"
                if "jobname" in self.inputs.run_opts.get_dict()['options']:
                    subname = subname + "-" +\
                        self.inputs.run_opts.get_dict()['options']['jobname']

                data_folder = repo.get_subfolder(subname)
                posout_list = data_folder.get_content_list(pattern="posout*")
                if posout_list.empty():
                    # not even, we failed. Should have been caught before.
                    self.report(
                        'Relaxation failed - no output found')
                    return self.exit_codes.ERROR_FAILED_RELAX
                sf = repo.get_abs_path(posout_list.sort()[-1][0])
        else:
            # no relaxation performed, file is named forces_posinp.xyz .. or yaml

            outstruct = "forces_posinp." + extension
            repo = workchain.outputs.retrieved._repository._get_base_folder()
            if "jobname" in self.inputs.run_opts.get_dict()['options']:
                outstruct = "forces_" +\
                    self.inputs.run_opts.get_dict(
                    )['options']['jobname'] + ".xyz"
            try:
                sf = repo.get_abs_path(outstruct, check_existence=True)
            except OSError:
                self.report(
                    'Relaxation failed - no output found')
                return self.exit_codes.ERROR_FAILED_RELAX

        s = StructureData()
        # BigDFT xyz files have more data on the first line which confuse aiida
        # Get rid of them (and get energy and forces in the process)

        if extension == "xyz":
            try:
                with open(sf) as f:
                    content = f.readlines()
            except FileNotFoundError:
                self.report(
                    'Relaxation failed - no output position file found')
                return self.exit_codes.ERROR_FAILED_RELAX
            firstline = content[0].split()
            content[0] = firstline[0] + '\n'
            self.out('total_energy', orm.Float(firstline[2]).store())

            forces_index = content.index(" forces\n")
            forces = content[forces_index + 1:]
            for i, f in enumerate(forces):
                forces[i] = f.split()[1:4]

            array_forces = ArrayData()
            array_forces.set_array('forces', numpy.array(forces))
            self.out('forces', array_forces.store())

            positions = content[0:forces_index]
            s._parse_xyz("".join(positions))
            s._adjust_default_cell(vacuum_addition=0.0,
                                   pbc=self.inputs.structure.pbc)
        elif extension == "yaml":
            logfile = workchain.outputs.bigdft_logfile.logfile
            if(isinstance(logfile, list)):
                energy = logfile[-1].get('Energy (Hartree)')
            else:
                energy = logfile.get('Energy (Hartree)')
            self.out('total_energy', orm.Float(energy).store())
            if(isinstance(logfile, list)):
                forces = logfile[-1].get('Atomic Forces (Ha/Bohr)')
            else:
                forces = logfile.get('Atomic Forces (Ha/Bohr)')
            for i, f in enumerate(forces):
                forces[i] = list(f.values())[0]
            array_forces = ArrayData()
            array_forces.set_array('forces', numpy.array(forces))
            self.out('forces', array_forces.store())
            logl = YamlIO.load(sf)
            if logl is None:
                self.report(
                    'Relaxation failed - no output position file found')
                return self.exit_codes.ERROR_FAILED_RELAX
            log = logl[0]
            s.cell = log['abc']
            for i in log['positions']:
                s.append_atom(symbols=list(i.keys())[
                              0], position=list(i.values())[0])
        else:
            self.report(
                'Relaxation failed - no parsing available for output type ' + extension)
            return self.exit_codes.ERROR_FAILED_RELAX
        self.out('relaxed_structure', s.store())
        self.out_many(self.exposed_outputs(workchain, BigDFTBaseWorkChain))



"""Handling of the input options

This module contains the useful quantities to deal with the preparation and
the usage of inputfiles for BigDFT code. The main object is the
:class:`Inputfile` class, which inherits from a python dictionary. Such
inheritance is made possible by the internal representation of the BigDFT
inputfile, which employs the YAML syntax. This means that there is a one-to-one
correspondence between a python dictionary and a BigDFT inputfile.

"""


class Inputfile(dict):
    """ The BigDFT inputfile.

    Principal object needed to run a  BigDFT calculation.
    Might be initialized either from a dictionary, a
    :py:class:`~BigDFT.Logfiles.Logfile` instance
    or a (yaml-compliant) filename path

    Note:

       Each of the actions of the module :py:mod:`~BigDFT.InputActions`, which
       is defined on a generic dictionary, also corresponds to a method of of
       the `Inputfile` class, and it is applied to the class instance.
       Therefore also the first argument of the corresponding action is
       implicitly the class instance. For the
       :py:func:`~BigDFT.InputActions.remove` method, the action has to be
       invoked should come from the :py:mod:`~BigDFT.InputActions` module.


    .. _input_action_example:
    Example:

       >>> import InputActions as A, Inputfiles as I
       >>> inp=I.Inputfile()
       >>> inp.set_hgrids(0.3) # equivalent to A.set_hgrids(inp,0.3)
       >>> inp
       {'dft': {'hgrids': 0.3}}
       >>> inp.optimize_geometry() # equivalent to A.optimize_geometry(inp)
       >>> inp
       {'dft': {'hgrids': 0.3},'geopt': {'method': 'FIRE',
                                         'ncount_cluster_x': 50} }
       >>> # equivalent to A.remove(inp,A.optimize_geometry)
       >>> inp.remove(A.optimize_geometry)
       >>> inp
       {'dft': {'hgrids': 0.3}}

     .. todo ::

         Consider the possiblity of initializing an `Inputfile` instance
         from a ``yaml`` file. And also from a
         :py:class:`~BigDFT.Logfiles.Logfile` class


    """

    def __init__(self, *args, **kwargs):
        import BigDFT.InputActions as A
        dict.__init__(self, *args, **kwargs)
        functions = dir(A)
        for action in functions:
            if "__" in action:
                continue
            from functools import partial
            func = getattr(A, action)
            setattr(self, action, partial(func, self))
